// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

//
// =======================================================================
//
// In the example below, `log_census`, `produce_cooling`, and `consume_cooling`
// are all plugins.  Their adh_plugin_definitions can be defined using these
// helper macros.
//
// census_plugin.c:
//
// DEFINE_LOGGER_PLUGIN(log_census, "log-census", log_census_create,
//      log_census_destroy, log_census_log);
//
// Define log_census_* functions listed above.
//
// cooling_plugin.c:
//
// DEFINE_PRODUCER_PLUGIN(produce_cooling, "produce-cooling",
// produce_cooling_create,
//      produce_cooling_destroy, produce_cooling_produce_sample,
//      produce_cooling_free_sample)
//
// DEFINE_CONSUMER_PLUGIN(consume_cooling, "consume-cooling",
// consume_cooling_create,
//      consume_cooling_destroy, consume_cooling_list_consume_targets,
//      consume_cooling_consume_sample)
//
// Define produce_cooling_* functions and consume_cooling_* functions.
//
// =============================
//
//
// Azure Device Health plugins need to be hosted inside a dynamic library
// that exposes a single function.  That function needs to return a
// NULL-terminated, static lifetime list of adh_plugin_definition pointers:
//
//
// module.c:
//
// extern const struct adh_plugin_definition log_census,
// produce_cooling_consume_cooling; extern const struct adh_plugin_definition
// produce_cooling; extern const struct adh_plugin_definition consume_cooling;
//
// static const adh_plugin_definition* const s_pluginsList[] = {
//     &log_census,
//     &produce_cooling,
//     &consume_cooling,
//     NULL
// };
//
// // NOTE: If you're using C++, make sure this is extern "C", and that
// //    this function has default symbol visibility.
//
// const adh_plugin_definition* const*
// get_adh_plugin_definitions()
// {
//     return s_pluginsList;
// }
//

struct adh_plugin_instance;

struct adh_plugin_config_pair
{
    const char* name;
    const char* value;
};

// Universal lifetime functions
typedef adh_plugin_instance* (*adh_create_plugin_instance_fn_ptr)(
    const struct adh_plugin_config_pair* configs,
    size_t configs_count);

typedef void (*adh_destroy_plugin_instance_fn_ptr)(
    struct adh_plugin_instance* plugin);

// Producer functions
struct adh_produce_sample_args
{
    uint64_t sample_key;
};
typedef void* (*adh_produce_sample_fn_ptr)(
    struct adh_plugin_instance* plugin,
    const struct adh_produce_sample_args* args);

typedef void (
    *adh_free_sample_fn_ptr)(struct adh_plugin_instance* plugin, void* sample);

// Consumer functions
typedef const char* const* (*adh_list_consume_targets_fn_ptr)();

struct adh_consume_sample_args
{
    uint64_t sample_key;
    const char* producer_name;
    void* sample;
};
typedef bool (*adh_consume_sample_fn_ptr)(
    struct adh_plugin_instance* plugin,
    const struct adh_consume_sample_args* args);

// Logger Functions
struct adh_log_data_args
{
    uint64_t sample_key;
};
typedef bool (*adh_log_data_fn_ptr)(
    struct adh_plugin_instance* plugin,
    const struct adh_log_data_args* args);

// Final struct definition used in exported function
struct adh_plugin_definition
{
    // Needed for all plugins
    const char* name;
    adh_create_plugin_instance_fn_ptr create_plugin_instance_fn;
    adh_destroy_plugin_instance_fn_ptr destroy_plugin_instance_fn;

    // Only needed for Produce plugins
    adh_produce_sample_fn_ptr produce_sample_fn;
    adh_free_sample_fn_ptr free_sample_fn;

    // Only needed for Consume plugins
    adh_list_consume_targets_fn_ptr list_consume_targets_fn;
    adh_consume_sample_fn_ptr consume_sample_fn;

    // Only needed for Log plugins
    adh_log_data_fn_ptr log_data_fn;
};

#define DEFINE_PRODUCER_PLUGIN(                                   \
    identifier, name, create, destroy, produceSample, freeSample) \
    extern const adh_plugin_definition identifier = {             \
        name,        &create, &destroy, &produceSample,           \
        &freeSample, nullptr, nullptr,  nullptr                   \
    };

#define DEFINE_CONSUMER_PLUGIN(                                           \
    identifier, name, create, destroy, listConsumeTargets, consumeSample) \
    extern const adh_plugin_definition identifier = {                     \
        name,           &create, &destroy,                                \
        nullptr,        nullptr, &listConsumeTargets,                     \
        &consumeSample, nullptr                                           \
    };

#define DEFINE_LOGGER_PLUGIN(identifier, name, create, destroy, logData)      \
    extern const adh_plugin_definition identifier = {                         \
        name, &create, &destroy, nullptr, nullptr, nullptr, nullptr, &logData \
    };
